@@ -8,6 +8,7 @@ from typedpy.structures import (
     TypedField,
     ClassReference,
     ImmutableStructure,
+    create_typed_field,
 )
 from typedpy.fields import (
     Number,
@@ -33,10 +34,16 @@ from typedpy.fields import (
     Tuple,
     StructureReference,
     ImmutableField,
-    create_typed_field,
     Anything,
     SerializableField,
     Function,
+    ImmutableMap,
+    ImmutableArray,
+    ImmutableSet,
+    ImmutableFloat,
+    ImmutableString,
+    ImmutableInteger,
+    ImmutableNumber,
 )
 
 from typedpy.json_schema_mapping import (
