@@ -3,23 +3,25 @@ Definitions of various types of fields. Supports JSON draft4 types.
 """
 import enum
 import re
-from abc import ABC
 from collections import OrderedDict
 from copy import deepcopy
 from functools import reduce
 from decimal import Decimal, InvalidOperation
 
+from typedpy.commons import wrap_val
 from typedpy.structures import (
     Field,
     Structure,
     TypedField,
     ClassReference,
     StructMeta,
-    is_function_returning_field,
+    ImmutableMixin,
+    _FieldMeta,
 )
 
 
 def _map_to_field(item):
+    item = item[0] if isinstance(item, (list, tuple)) and len(item) == 1 else item
     if isinstance(item, StructMeta) and not isinstance(item, Field):
         return ClassReference(item)
     if item in [None, ""] or isinstance(item, Field):
@@ -30,10 +32,6 @@ def _map_to_field(item):
         raise TypeError("Expected a Field/Structure class or Field instance")
 
 
-def wrap_val(v):
-    return "'{}'".format(v) if isinstance(v, str) else v
-
-
 class StructureReference(Field):
     """
     A Field that is an embedded structure within other structure. Allows to create hierarchy.
@@ -349,7 +347,7 @@ class PositiveInt(Integer, Positive):
     pass
 
 
-class _ListStruct(list):
+class _ListStruct(list, ImmutableMixin):
     """
     This is a useful wrapper for the content of list in an Array field.
     It ensures that an update of the form:
@@ -357,23 +355,46 @@ class _ListStruct(list):
     Will not bypass the validation of the Array.
     """
 
+    class ListIteratorProxy:
+        def __init__(self, the_list):
+            self.the_list = the_list
+            self.index = 0
+
+        def __next__(self):
+            if len(self.the_list) > self.index:
+                self.index += 1
+                return self.the_list[self.index - 1]
+            raise StopIteration
+
     def __init__(self, array: Field, struct_instance: Structure, mylist):
         self._field_definition = array
         self._instance = struct_instance
-        super().__init__(mylist)
+        super().__init__(self._get_defensive_copy_if_needed(mylist))
 
     def __setitem__(self, key, value):
+        self._raise_if_immutable()
         copied = self[:]
         copied.__setitem__(key, value)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
 
+    def __getitem__(self, item):
+        val = super().__getitem__(item)
+        return self._get_defensive_copy_if_needed(val)
+
+    def __iter__(self):
+        if self._is_immutable():
+            return _ListStruct.ListIteratorProxy(self)
+        return super(_ListStruct, self).__iter__()
+
     def append(self, value):
+        self._raise_if_immutable()
         copied = self[:]
         copied.append(value)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
         super().append(value)
 
     def extend(self, value):
+        self._raise_if_immutable()
         copied = self[:]
         copied.extend(value)
         if getattr(self, "_instance", None):
@@ -382,16 +403,27 @@ class _ListStruct(list):
             )
 
     def insert(self, index: int, value):
+        self._raise_if_immutable()
         copied = self[:]
         copied.insert(index, value)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
 
     def remove(self, ind):
+        self._raise_if_immutable()
         copied = self[:]
         copied.remove(ind)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
 
+    def copy(self):
+        copied = super(_ListStruct, self).copy()
+        return deepcopy(copied) if self._is_immutable() else copied
+
+    def clear(self) -> None:
+        self._raise_if_immutable()
+        setattr(self._instance, getattr(self._field_definition, "_name", None), [])
+
     def pop(self, index: int = -1):
+        self._raise_if_immutable()
         copied = self[:]
         res = copied.pop(index)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
@@ -418,7 +450,7 @@ class _ListStruct(list):
         super().__init__(state["the_values"])
 
 
-class _DictStruct(dict):
+class _DictStruct(dict, ImmutableMixin):
     """
     This is a useful wrapper for the content of dict in an Map field.
     It ensures that an update of the form:
@@ -434,25 +466,60 @@ class _DictStruct(dict):
         super().__init__(mydict)
 
     def __setitem__(self, key, value):
+        super()._raise_if_immutable()
         copied = self.copy()
         copied.__setitem__(key, value)
         if getattr(self, "_instance", None):
             setattr(
                 self._instance, getattr(self._field_definition, "_name", None), copied
             )
+
         super().__setitem__(key, value)
 
+    def __getitem__(self, item):
+        val = super().__getitem__(item)
+        return self._get_defensive_copy_if_needed(val)
+
+    def copy(self):
+        copied = super(_DictStruct, self).copy()
+        return deepcopy(copied) if self._is_immutable() else copied
+
+    def items(self):
+        return (
+            (k, self._get_defensive_copy_if_needed(v))
+            for k, v in super(_DictStruct, self).items()
+        )
+
+    def values(self):
+        return (
+            self._get_defensive_copy_if_needed(v)
+            for v in super(_DictStruct, self).values()
+        )
+
     def __delitem__(self, key):
+        self._raise_if_immutable()
         copied = self.copy()
         del copied[key]
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
 
     def update(self, *args, **kwargs):
+        self._raise_if_immutable()
         copied = self.copy()
         res = copied.update(*args, **kwargs)
         setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
         return res
 
+    def pop(self, k):
+        self._raise_if_immutable()
+        copied = self.copy()
+        res = copied.pop(k)
+        setattr(self._instance, getattr(self._field_definition, "_name", None), copied)
+        return res
+
+    def clear(self) -> None:
+        self._raise_if_immutable()
+        setattr(self._instance, getattr(self._field_definition, "_name", None), {})
+
     def __getstate__(self):
         return {
             "_instance": self._instance,
@@ -466,19 +533,10 @@ class _DictStruct(dict):
         super().__init__(state["mydict"])
 
 
-class _CollectionMeta(type):
+class _CollectionMeta(_FieldMeta):
     def __getitem__(cls, item):
         def validate_and_get_field(val):
-            if isinstance(val, Field):
-                return val
-            elif Field in getattr(val, "__mro__", {}):
-                return val()
-            elif Structure in getattr(val, "__mro__", {}):
-                return ClassReference(val)
-            elif is_function_returning_field(val):
-                return val()
-            else:
-                raise TypeError("Expected a Field class or instance")
+            return _FieldMeta.__getitem__(cls, val)
 
         if isinstance(item, tuple):
             items = [validate_and_get_field(it) for it in item]
@@ -486,26 +544,17 @@ class _CollectionMeta(type):
         return cls(items=validate_and_get_field(item))
 
 
-class _EnumMeta(type):
+class _EnumMeta(_FieldMeta):
     def __getitem__(cls, values):
         if isinstance(values, (type,)) and issubclass(values, (enum.Enum,)):
             return cls(values=values)
         return cls(values=list(values))
 
 
-class _JSONSchemaDraft4ReuseMeta(type):
+class _JSONSchemaDraft4ReuseMeta(_FieldMeta):
     def __getitem__(cls, item):
         def validate_and_get_field(val):
-            if isinstance(val, Field):
-                return val
-            elif Field in getattr(val, "__mro__", {}):
-                return val()
-            elif Structure in getattr(val, "__mro__", {}):
-                return ClassReference(val)
-            elif is_function_returning_field(val):
-                return val()
-            else:
-                raise TypeError("Expected a Field class or instance")
+            return _FieldMeta.__getitem__(cls, val)
 
         if isinstance(item, tuple):
             fields = [validate_and_get_field(it) for it in item]
@@ -576,19 +625,20 @@ class Set(SizedCollection, TypedField, metaclass=_CollectionMeta):
         super().__init__(*args, **kwargs)
 
     def __set__(self, instance, value):
-        if not isinstance(value, set):
+        cls = self.__class__._ty
+        if not isinstance(value, cls):
             raise TypeError(
-                "{}: Got {}; Expected {}".format(self._name, wrap_val(value), set)
+                "{}: Got {}; Expected {}".format(self._name, wrap_val(value), cls)
             )
         self.validate_size(value, self._name)
         if self.items is not None:
             temp_st = Structure()
             setattr(self.items, "_name", self._name)
-            res = set()
+            res = []
             for val in value:
                 self.items.__set__(temp_st, val)
-                res.add(getattr(temp_st, getattr(self.items, "_name")))
-                value = res
+                res.append(getattr(temp_st, getattr(self.items, "_name")))
+            value = cls(res)
         super().__set__(instance, value)
 
 
@@ -659,7 +709,6 @@ class Map(SizedCollection, TypedField, metaclass=_CollectionMeta):
                 res[getattr(temp_st, getattr(key_field, "_name"))] = getattr(
                     temp_st, getattr(value_field, "_name")
                 )
-            value = res
         super().__set__(instance, _DictStruct(self, instance, value))
 
 
@@ -1160,53 +1209,52 @@ class NotField(MultiFieldWrapper, Field, metaclass=_JSONSchemaDraft4ReuseMeta):
         return _str_for_multioption_field(self)
 
 
-class ValidatedTypedField(TypedField):
+class ImmutableSet(Set, ImmutableField):
+    _ty = frozenset
+
     def __set__(self, instance, value):
-        self._validate_func(value)  # pylint: disable=E1101
-        super().__set__(instance, value)
+        if not isinstance(value, (set, frozenset)):
+            raise TypeError(
+                "{}: Got {}; Expected {}".format(self._name, wrap_val(value), set)
+            )
+        self.validate_size(value, self._name)
+        if self.items is not None:
+            temp_st = Structure()
+            setattr(self.items, "_name", self._name)
+            res = set()
+            for val in value:
+                self.items.__set__(temp_st, val)
+                res.add(getattr(temp_st, getattr(self.items, "_name")))
+                value = res
+        corrected_value = value if isinstance(value, frozenset) else frozenset(value)
+        super().__set__(instance, corrected_value)
 
 
-def create_typed_field(classname, cls, validate_func=None):
-    """
-    Factory that generates a new class for a :class:`Field` as a wrapper of any class.
-    Example:
-    Given a class Foo, and a validation function for the value in Foo - validate_foo, the line
-
-    .. code-block:: python
-
-        ValidatedFooField = create_typed_field("FooField", Foo, validate_func=validate_foo)
-
-    Generates a new :class:`Field` class that validates the content using validate_foo, and can be
-    used just like any :class:`Field` type.
-
-    .. code-block:: python
-
-        class A(Structure):
-            foo = ValidatedFooField
-            bar = Integer
-
-        # asumming we have an instance of Foo, called my_foo:
-        A(bar=4, foo=my_foo)
-
-    Arguments:
-
-        classname(`str`):
-            the content must not match any of the fields in the lists
-    """
-
-    def validate_wrapper(cls, value):
-        if validate_func is None:
-            return
-        validate_func(value)
-
-    return type(
-        classname,
-        (ValidatedTypedField,),
-        {"_validate_func": validate_wrapper, "_ty": cls},
-    )
+class ImmutableMap(ImmutableField, Map):
+    pass
 
 
-class SerializableField(ABC):
+class ImmutableArray(ImmutableField, Array):
+    pass
+
+
+class ImmutableString(ImmutableField, String):
+    pass
+
+
+class ImmutableNumber(ImmutableField, Number):
+    pass
+
+
+class ImmutableInteger(ImmutableField, Integer):
+    pass
+
+
+class ImmutableFloat(ImmutableField, Float):
+    pass
+
+
+class SerializableField(Field):
     """
     An abstract class for a field that has custom serialization or deserialization.
     can override the method:
@@ -1232,4 +1280,4 @@ class ExceptionField(TypedField, SerializableField):
     _ty = Exception
 
     def serialize(self, value):
-        return str(value)
+        return "{}: {}".format(value.__class__.__name__, str(value))
