@@ -10,7 +10,7 @@ from typedpy.structures import (
     Structure,
     _get_all_fields_by_name,
     ADDITIONAL_PROPERTIES,
-    REQUIRED,
+    REQUIRED_FIELDS,
 )
 from typedpy.fields import (
     Field,
@@ -201,9 +201,15 @@ def deserialize_single_field(
         value = field.deserialize(source_val)
     elif isinstance(field, Anything) or field is None:
         value = source_val
+    elif isinstance(field, TypedField) and isinstance(source_val, (list, dict)):
+        ty = getattr(field, "_ty")
+        if isinstance(source_val, list):
+            value = ty(*source_val)
+        elif isinstance(source_val, dict):
+            value = ty(**source_val)
     else:
         raise NotImplementedError(
-            "cannot deserialize field '{}' of type {} using value {}".format(
+            "cannot deserialize field '{}' of type {} using value {}. Are you using non-Typepy class?".format(
                 name, field.__class__.__name__, wrap_val(source_val)
             )
         )
@@ -300,7 +306,7 @@ def deserialize_structure_internal(
     if not isinstance(the_dict, dict):
         props = cls.__dict__
         fields = list(field_by_name.keys())
-        required = props.get(REQUIRED, fields)
+        required = props.get(REQUIRED_FIELDS, fields)
         additional_props = props.get(ADDITIONAL_PROPERTIES, True)
         if len(fields) == 1 and required == fields and additional_props is False:
             field_name = fields[0]
@@ -492,7 +498,7 @@ def serialize_internal(structure, mapper=None, compact=False):
     additional_props = props.get(ADDITIONAL_PROPERTIES, True)
     if (
         len(fields) == 1
-        and props.get(REQUIRED, fields) == fields
+        and props.get(REQUIRED_FIELDS, fields) == fields
         and additional_props is False
         and compact
     ):
