@@ -2,17 +2,40 @@
 The Skeleton classes to support strictly defined structures:
 Structure, Field, StructureReference, ClassReference, TypedField
 """
+import enum
 from copy import deepcopy
 from collections import OrderedDict
-from inspect import Signature, Parameter
+from inspect import Signature, Parameter, signature
 import sys
+import typing
+import hashlib
 
 from typing import get_type_hints, Iterable
 
-REQUIRED = "_required"
+from typedpy.commons import wrap_val
+
+REQUIRED_FIELDS = "_required"
 DEFAULTS = "_defaults"
 ADDITIONAL_PROPERTIES = "_additionalProperties"
 IS_IMMUTABLE = "_immutable"
+OPTIONAL_FIELDS = "_optional"
+
+
+class ImmutableMixin:
+    _field_definition = None
+    _instance = None
+
+    def _get_defensive_copy_if_needed(self, value):
+        return deepcopy(value) if self._is_immutable() else value
+
+    def _is_immutable(self):
+        instance = getattr(self, "_instance", None)
+        return getattr(self._instance, IS_IMMUTABLE, False) if instance else False
+
+    def _raise_if_immutable(self):
+        if self._is_immutable():
+            name = self._field_definition, "_name", None
+            raise ValueError("{}: Structure is immutable".format(name))
 
 
 def make_signature(names, required, additional_properties, bases_params_by_name):
@@ -92,7 +115,41 @@ def get_base_info(bases):
     return bases_params, bases_required
 
 
-class Field:
+class _FieldMeta(type):
+    _registry = {}
+
+    def __getitem__(cls, val):
+        if isinstance(val, Field):
+            return val
+        elif Field in getattr(val, "__mro__", {}):
+            return val()
+        elif Structure in getattr(val, "__mro__", {}):
+            return ClassReference(val)
+        elif is_function_returning_field(val):
+            return val()
+        else:
+
+            def get_state(value):
+                raise TypeError(
+                    "pickling of implicit wrappers for non-Typedpy fields are unsupported"
+                )
+
+            if not isinstance(val, type):
+                raise TypeError(
+                    "Unsupported field type in definition: {}".format(wrap_val(val))
+                )
+            the_class = val.__name__
+            if the_class in _FieldMeta._registry:
+                return _FieldMeta._registry[the_class]
+            short_hash = hashlib.sha256(the_class.encode("utf-8")).hexdigest()[:8]
+            new_name = "Field_{}_{}".format(the_class, short_hash)
+            class_as_field = create_typed_field(new_name, val)
+            class_as_field.__getstate__ = get_state
+            _FieldMeta._registry[the_class] = class_as_field
+            return class_as_field()
+
+
+class Field(metaclass=_FieldMeta):
     """
     Base class for a field(i.e. property) in a structure.
     Should not be used directly by developers.
@@ -111,17 +168,29 @@ class Field:
             self.__set__(Structure(), default)
         except Exception as e:
             raise e.__class__(
-                "Invalid default value: {}; Reason: {}".format(default, str(e))
+                "Invalid default value: {}; Reason: {}".format(
+                    wrap_val(default), str(e)
+                )
             ) from e
 
     def __get__(self, instance, owner):
         if instance is not None and self._name not in instance.__dict__:
             return self._default
-        return (
+        res = (
             instance.__dict__[self._name]
             if instance is not None
             else owner.__dict__[self._name]
         )
+        is_immutable = (
+            instance is not None
+            and getattr(instance, IS_IMMUTABLE, False)
+            or getattr(self, IS_IMMUTABLE, False)
+        )
+        needs_defensive_copy = (
+            not isinstance(res, (ImmutableMixin, int, float, str, bool, enum.Enum))
+            or res is None
+        )
+        return deepcopy(res) if (is_immutable and needs_defensive_copy) else res
 
     def __set__(self, instance, value):
         if getattr(self, IS_IMMUTABLE, False) and self._name in instance.__dict__:
@@ -167,6 +236,8 @@ def is_function_returning_field(field_definition_candidate):
     python_ver_higher_than_36 = sys.version_info[0:2] != (3, 6)
     if callable(field_definition_candidate) and python_ver_higher_than_36:
         try:
+            if len(signature(field_definition_candidate).parameters) > 0:
+                raise TypeError("function not allowed to accept any parameters")
             return_value = get_type_hints(field_definition_candidate).get(
                 "return", None
             )
@@ -192,8 +263,16 @@ def _get_all_fields_by_name(cls):
 def _instantiate_fields_if_needed(cls_dict: dict, defaults: dict):
     for key, val in cls_dict.items():
         if (
-            key not in {REQUIRED, ADDITIONAL_PROPERTIES, IS_IMMUTABLE, DEFAULTS}
+            key
+            not in {
+                REQUIRED_FIELDS,
+                ADDITIONAL_PROPERTIES,
+                IS_IMMUTABLE,
+                DEFAULTS,
+                OPTIONAL_FIELDS,
+            }
             and not isinstance(val, Field)
+            and not key.startswith("__")
             and (
                 Field in getattr(val, "__mro__", []) or is_function_returning_field(val)
             )
@@ -205,20 +284,22 @@ def _instantiate_fields_if_needed(cls_dict: dict, defaults: dict):
 def _apply_default_and_update_required_not_to_include_fields_with_defaults(
     cls_dict: dict, defaults: dict, fields: list
 ):
-    required_fields = set(cls_dict.get(REQUIRED, []))
-    required_fields_predefined = REQUIRED in cls_dict
+    required_fields = set(cls_dict.get(REQUIRED_FIELDS, []))
+    optional_fields = set(cls_dict.get(OPTIONAL_FIELDS, []))
+    required_fields_predefined = REQUIRED_FIELDS in cls_dict
     for field_name in fields:
         if field_name in defaults and not getattr(
             cls_dict[field_name], "_default", None
         ):
             cls_dict[field_name]._try_default_value(defaults[field_name])
             cls_dict[field_name]._default = defaults[field_name]
-        if getattr(cls_dict[field_name], "_default", None):
+        if getattr(cls_dict[field_name], "_default", None) is not None:
             if field_name in required_fields:
                 required_fields.remove(field_name)
         elif not required_fields_predefined:
-            required_fields.add(field_name)
-    cls_dict[REQUIRED] = list(required_fields)
+            if field_name not in optional_fields:
+                required_fields.add(field_name)
+    cls_dict[REQUIRED_FIELDS] = list(required_fields)
 
 
 class StructMeta(type):
@@ -256,7 +337,7 @@ class StructMeta(type):
         default_required = (
             list(set(bases_required + fields)) if bases_params else fields
         )
-        required = cls_dict.get(REQUIRED, default_required)
+        required = cls_dict.get(REQUIRED_FIELDS, default_required)
         additional_props = cls_dict.get(ADDITIONAL_PROPERTIES, True)
         sig = make_signature(clsobj._fields, required, additional_props, bases_params)
         setattr(clsobj, "__signature__", sig)
@@ -272,6 +353,79 @@ class StructMeta(type):
         return "<Structure: {}. Properties: {}>".format(name, ", ".join(props))
 
 
+def convert_basic_types(v):
+    from .fields import (
+        Integer,
+        Float,
+        String,
+        Map,
+        Array,
+        Tuple,
+        Set,
+        Boolean,
+        ImmutableSet,
+        AnyOf,
+        Anything,
+    )
+
+    type_mapping = {
+        int: Integer,
+        str: String,
+        float: Float,
+        dict: Map,
+        set: Set,
+        list: Array,
+        tuple: Tuple,
+        bool: Boolean,
+        frozenset: ImmutableSet,
+        typing.Union: AnyOf,
+        typing.Any: Anything,
+    }
+    return type_mapping.get(v, None)
+
+
+def get_typing_lib_info(v):
+    from .fields import AnyOf
+
+    class Foo:
+        pass
+
+    py_version = sys.version_info[0:2]
+    python_ver_atleast_than_37 = py_version >= (3, 6)
+    python_ver_atleast_39 = py_version >= (3, 9)
+
+    generic_alias = getattr(typing, "_GenericAlias", Foo)
+    special_generic_alias = getattr(typing, "_SpecialGenericAlias", Foo)
+    origin = getattr(v, "__origin__", None)
+    is_typing_generic = (
+        python_ver_atleast_than_37
+        and isinstance(v, (generic_alias, special_generic_alias))
+    ) or (
+        python_ver_atleast_39
+        and origin in {list, dict, tuple, set, frozenset, typing.Union}
+    )
+    if not is_typing_generic:
+        return convert_basic_types(v)
+    mapped_type = convert_basic_types(origin)
+    if mapped_type is None:
+        raise TypeError("{} type is not supported".format(v))
+    args_raw = getattr(v, "__args__", None)
+    if not args_raw:
+        return mapped_type()
+    mapped_args = [
+        get_typing_lib_info(a) for a in args_raw if not isinstance(a, typing.TypeVar)
+    ]
+    if not all(mapped_args):
+        raise TypeError("invalid type {}".format(v))
+    if mapped_args:
+        if mapped_type == AnyOf:
+            return mapped_type(fields=mapped_args)
+        else:
+            mapped_args = mapped_args if len(mapped_args) > 1 else mapped_args[0]
+            return mapped_type(items=mapped_args)
+    return mapped_type()
+
+
 def add_annotations_to_class_dict(cls_dict):
     annotations = cls_dict.get("__annotations__", {})
     defaults = {}
@@ -283,33 +437,16 @@ def add_annotations_to_class_dict(cls_dict):
                 if k in cls_dict:
                     defaults[k] = cls_dict[k]
                 cls_dict[k] = v
-            elif v in {int, float, str, dict, list, tuple, set}:
-                from .fields import (
-                    Integer,
-                    Float,
-                    String,
-                    Map,
-                    Array,
-                    Tuple,
-                    Set,
-                    Boolean,
-                )
-
-                type_mapping = {
-                    int: Integer,
-                    str: String,
-                    float: Float,
-                    dict: Map,
-                    set: Set,
-                    list: Array,
-                    tuple: Tuple,
-                    bool: Boolean,
-                }
-                the_type = type_mapping[v]
-                if k in cls_dict:
-                    the_type = the_type(default=cls_dict[k])
-                    defaults[k] = cls_dict[k]
-                cls_dict[k] = the_type
+            else:
+                the_type = get_typing_lib_info(v)
+                if the_type:
+                    if k in cls_dict:
+                        try:
+                            the_type = the_type(default=cls_dict[k])
+                        except Exception as e:
+                            raise e.__class__("{}: {}".format(k, str(e))) from e
+                        defaults[k] = cls_dict[k]
+                    cls_dict[k] = the_type
     cls_dict[DEFAULTS] = defaults
 
 
@@ -338,6 +475,18 @@ class Structure(metaclass=StructMeta):
                 # this raises an exception:
                 Foo(name="John")
 
+        _optional: optional
+            If we don't state the "_required" field, we can state which fields are optional instead.
+            Example:
+
+            .. code-block:: python
+
+                class Foo(Structure):
+                    id = Integer
+                    name = String
+                    _optional = ['name']
+
+
         _additionalProperties(bool): optional
             Is it allowed to add additional properties that are not defined in the class definition?
             the default is True.
@@ -441,7 +590,9 @@ class Structure(metaclass=StructMeta):
         return str(self).__hash__()
 
     def __delitem__(self, key):
-        if isinstance(getattr(self, REQUIRED), list) and key in getattr(self, REQUIRED):
+        if isinstance(getattr(self, REQUIRED_FIELDS), list) and key in getattr(
+            self, REQUIRED_FIELDS
+        ):
             raise ValueError("{} is mandatory".format(key))
         del self.__dict__[key]
 
@@ -482,7 +633,7 @@ class Structure(metaclass=StructMeta):
         field_by_name = _get_all_fields_by_name(self.__class__)
         field_names = list(field_by_name.keys())
         props = self.__class__.__dict__
-        required = props.get(REQUIRED, field_names)
+        required = props.get(REQUIRED_FIELDS, field_names)
         additional_props = props.get(ADDITIONAL_PROPERTIES, True)
         if (
             len(field_names) == 1
@@ -533,8 +684,10 @@ class TypedField(Field):
         def err_prefix():
             return "{}: ".format(self._name) if self._name else ""
 
-        if not isinstance(value, (self._ty)) and value is not None:
-            raise TypeError("{}Expected {}".format(err_prefix(), self._ty))
+        if not isinstance(value, self._ty) and value is not None:
+            raise TypeError(
+                "{}Expected {}; Got {}".format(err_prefix(), self._ty, wrap_val(value))
+            )
 
     def __set__(self, instance, value):
         if not getattr(instance, "_skip_validation", False):
@@ -542,6 +695,52 @@ class TypedField(Field):
         super().__set__(instance, value)
 
 
+class ValidatedTypedField(TypedField):
+    def __set__(self, instance, value):
+        self._validate_func(value)  # pylint: disable=E1101
+        super().__set__(instance, value)
+
+
+def create_typed_field(classname, cls, validate_func=None):
+    """
+    Factory that generates a new class for a :class:`Field` as a wrapper of any class.
+    Example:
+    Given a class Foo, and a validation function for the value in Foo - validate_foo, the line
+
+    .. code-block:: python
+
+        ValidatedFooField = create_typed_field("FooField", Foo, validate_func=validate_foo)
+
+    Generates a new :class:`Field` class that validates the content using validate_foo, and can be
+    used just like any :class:`Field` type.
+
+    .. code-block:: python
+
+        class A(Structure):
+            foo = ValidatedFooField
+            bar = Integer
+
+        # asumming we have an instance of Foo, called my_foo:
+        A(bar=4, foo=my_foo)
+
+    Arguments:
+
+        classname(`str`):
+            the content must not match any of the fields in the lists
+    """
+
+    def validate_wrapper(cls, value):
+        if validate_func is None:
+            return
+        validate_func(value)
+
+    return type(
+        classname,
+        (ValidatedTypedField,),
+        {"_validate_func": validate_wrapper, "_ty": cls},
+    )
+
+
 class ClassReference(TypedField):
     """
     A field that is a reference to another Structure instance.
